@@ -5,7 +5,6 @@
 VALUE eXMLError;
 static ID CALL_METHOD;
 static ID ERROR_HANDLER_ID;
-static ID ON_ERROR_METHOD;
 
 /*
  * Document-class: LibXML::XML::Error
@@ -23,6 +22,10 @@ static ID ON_ERROR_METHOD;
  *
  *   XML::Error.set_handler(&XML::Error::QUIET_HANDLER)
  *
+ * Get the current handler:
+ *
+ *   proc = XML::Error.get_handler
+ *
  * Install your own handler:
  *
  *   XML::Error.set_handler do |error|
@@ -34,48 +37,17 @@ static ID ON_ERROR_METHOD;
  *   XML::Error.reset_handler
  */
 
-static void rxml_set_handler(VALUE self, VALUE block)
-{
-#ifdef RB_CVAR_SET_4ARGS
-  rb_cvar_set(self, ERROR_HANDLER_ID, block, 0);
-#else
-  rb_cvar_set(self, ERROR_HANDLER_ID, block);
-#endif
-}
-
 /*
  * call-seq:
- *    Error.set_error_handler {|error| ... }
+ *    Error.get_error_handler
  *
- * Registers a block that will be called with an instance of
- * XML::Error when libxml generates warning, error or fatal
- * error messages.
+ * Returns the proc that will be called when libxml generates
+ * warning, error or fatal error messages.
  */
-static VALUE rxml_error_set_handler(VALUE self)
+static VALUE rxml_error_get_handler(VALUE self)
 {
-  VALUE block;
-
-  if (rb_block_given_p() == Qfalse)
-    rb_raise(rb_eRuntimeError, "No block given");
-
-  block = rb_block_proc();
-
-  /* Embed the block within the Error class to avoid it to be collected.
-   Previous handler will be overwritten if it exists. */
-  rxml_set_handler(self, block);
-
-  return self;
-}
-
-/*
- * call-seq:
- *    Error.reset_error_handler
- *
- * Removes the current error handler. */
-static VALUE rxml_error_reset_handler(VALUE self)
-{
-  rxml_set_handler(self, Qnil);
-  return self;
+  VALUE block = rb_cvar_get(eXMLError, ERROR_HANDLER_ID);
+  return block;
 }
 
 VALUE rxml_error_wrap(xmlErrorPtr xerror)
@@ -130,31 +102,7 @@ static void structuredErrorFunc(void *userData, xmlErrorPtr xerror)
   VALUE error = rxml_error_wrap(xerror);
 
   /* Wrap error up as Ruby object and send it off to ruby */
-  VALUE block = rb_cvar_get(eXMLError, ERROR_HANDLER_ID);
-
-  /* This next bit of code is a total hack to get around a bug
-     in libxml which causes error handlers on sax handlers to
-     be ignored in favor of the global handler.  In addition,
-     the correct context is also not passed in.  So try to 
-     dig it out. */
-  if (!userData && xerror->ctxt)
-  {
-    xmlParserCtxtPtr ctxt = (xmlParserCtxtPtr) xerror->ctxt;
-    if (ctxt != ctxt->userData)
-      userData = ctxt->userData;
-  }
-
-  /* If the target has an on_error method call it.  This
-   gets around a bug in libxml where a sax's structured
-   error handler is overriden by the global error handler. */
-  if (userData)
-  {  
-    VALUE target = (VALUE) userData;
-    if (!NIL_P(target) && rb_respond_to(target, ON_ERROR_METHOD))
-    {
-      rb_funcall(target, ON_ERROR_METHOD, 1, error);
-    }
-  }
+  VALUE block = rxml_error_get_handler(error);
 
   /* Now call global handler */
   if (block != Qnil)
@@ -163,24 +111,68 @@ static void structuredErrorFunc(void *userData, xmlErrorPtr xerror)
   }
 }
 
-void rxml_raise(xmlErrorPtr xerror)
+static void rxml_set_handler(VALUE self, VALUE block)
+{
+#ifdef RB_CVAR_SET_4ARGS
+  rb_cvar_set(self, ERROR_HANDLER_ID, block, 0);
+#else
+  rb_cvar_set(self, ERROR_HANDLER_ID, block);
+#endif
+
+  /* Intercept libxml error handlers */
+  xmlSetStructuredErrorFunc(NULL, structuredErrorFunc);
+}
+
+/*
+ * call-seq:
+ *    Error.set_error_handler {|error| ... }
+ *
+ * Registers a block that will be called with an instance of
+ * XML::Error when libxml generates warning, error or fatal
+ * error messages.
+ */
+static VALUE rxml_error_set_handler(VALUE self)
+{
+  VALUE block;
+
+  if (rb_block_given_p() == Qfalse)
+    rb_raise(rb_eRuntimeError, "No block given");
+
+  block = rb_block_proc();
+
+  /* Embed the block within the Error class to avoid it to be collected.
+   Previous handler will be overwritten if it exists. */
+  rxml_set_handler(self, block);
+
+  return self;
+}
+
+/*
+ * call-seq:
+ *    Error.reset_error_handler
+ *
+ * Removes the current error handler. */
+static VALUE rxml_error_reset_handler(VALUE self)
+{
+  rxml_set_handler(self, Qnil);
+  return self;
+}
+
+NORETURN(void rxml_raise(xmlErrorPtr xerror))
 {
   /* Wrap error up as Ruby object and send it off to ruby */
   VALUE error = rxml_error_wrap(xerror);
   rb_exc_raise(error);
 }
 
-void rxml_init_error()
+void rxml_init_error(void)
 {
-  ON_ERROR_METHOD = rb_intern("on_error");
   CALL_METHOD = rb_intern("call");
   ERROR_HANDLER_ID = rb_intern("@@__error_handler_callback__");
 
-  /* Intercept libxml error handlers */
-  xmlSetStructuredErrorFunc(NULL, structuredErrorFunc);
-
   /* Error class */
   eXMLError = rb_define_class_under(mXML, "Error", rb_eStandardError);
+  rb_define_singleton_method(eXMLError, "get_handler", rxml_error_get_handler, 0);
   rb_define_singleton_method(eXMLError, "set_handler", rxml_error_set_handler, 0);
   rb_define_singleton_method(eXMLError, "reset_handler", rxml_error_reset_handler, 0);
 
