@@ -2,6 +2,10 @@
 #include "ruby_xml_node.h"
 #include <assert.h>
 
+#include <libxml/debugXML.h>
+#include <libxml/parserInternals.h>
+#include <libxml/xlink.h>
+
 VALUE cXMLNode;
 
 /* Document-class: LibXML::XML::Node
@@ -17,96 +21,127 @@ VALUE cXMLNode;
 
 /* Memory management:
  *
- * The bindings create a one-to-one mapping between libxml nodes
- * and Ruby nodes.  If a libxml node is wraped, its _private member
- * is set with a reference to the Ruby object.
+ * The bindings create a one-to-one mapping between ruby objects and
+ * libxml documents and libxml parent nodes (ie, nodes that do not
+ * have a parent and do not belong to a document). In these cases,
+ * the bindings manage the memory.  They do this by installing a free
+ * function and storing a back pointer to the Ruby object from the xmlnode
+ * using the _private member on libxml structures.  When the Ruby object
+ * goes out of scope, the underlying libxml structure is freed.  Libxml
+ * itself then frees all child node (recursively).
  *
- * When a libxml document or top level node is freed, it will free
- * all its children.  Thus Ruby is responsible for:
+ * For all other nodes (the vast majority), the bindings create temporary
+ * Ruby objects that get freed once they go out of scope. Thus there can be
+ * more than one ruby object pointing to the same xml node.  To mostly hide
+ * this from programmers on the ruby side, the #eql? and #== methods are 
+ * overriden to check if two ruby objects wrap the same xmlnode.  If they do,
+ * then the methods return true.  During the mark phase, each of these temporary
+ * objects marks its owning document, thereby keeping the Ruby document object
+ * alive and thus the xmldoc tree.
  *
- *  * Using the mark function to keep alive any documents Ruby is
- *    referencing via the document or child nodes.
- *  * Using the mark function to keep alive any top level, free
- *    standing nodes Ruby is referencing via the node or its children.
- *
- * In general use, this will cause Ruby nodes to be freed before
- * a libxml document.  When a Ruby node is freed, the _private
- * field is set back to null.  
- *
- * In the sweep phase in Ruby 1.9.1, the document is freed before
- * the nodes.  To support that, the bindingds register a callback 
- * function with libxml that is called each time a node is freed.
- * In that case, the data_ptr is set to null, so the bindings
- * can recognize the situtation.
+ * In the sweep phase of the garbage collector, or when a program ends, 
+ * there is no order to how Ruby objects are freed. In fact, the ruby document
+ * object is almost always freed before any ruby objects that wrap child nodes.
+ * However, this is ok because those ruby objects do not have a free function
+ * and are no longer in scope (since if they were the document would not be freed).
  */
 
-static void rxml_node_deregisterNode(xmlNodePtr xnode)
-{
-  /* Has the node been wrapped and exposed to Ruby? */
-  if (xnode->_private)
-  {
-    /* Node was wrapped.  Set the _private member to free and
-      then dislabe the dfree function so that Ruby will not
-      try to free the node a second time. */
-    VALUE node = (VALUE) xnode->_private;
-    RDATA(node)->data = NULL;
-  }
-}
-
 static void rxml_node_free(xmlNodePtr xnode)
 {
-  /* Either the node has been created yet in initialize
-     or it has been freed by libxml already in Ruby's 
-     mark phase. */
-  if (xnode == NULL)
-    return;
-
-  /* The ruby object wrapping the xml object no longer exists. */
-  xnode->_private = NULL;
-
-  /* Ruby is responsible for freeing this node since it
-     has no parent. */
-  if (xnode->parent == NULL)
+  /* The ruby object wrapping the xml object no longer exists and this
+     is a standalone node without a document or parent so ruby is 
+     responsible for freeing the underlying node.*/
+  if (xnode->doc == NULL && xnode->parent == NULL)
+  {
+    // Remove the back linkage from libxml to Ruby
+    xnode->_private = NULL;
     xmlFreeNode(xnode);
+  }
 }
 
- void rxml_node_mark(xmlNodePtr xnode)
+void rxml_node_manage(xmlNodePtr xnode, VALUE node)
 {
-  /* Either the node has been created yet in initialize
-     or it has been freed by libxml already in Ruby's 
-     mark phase. */
-  if (xnode == NULL)
-    return;
+  RDATA(node)->dfree = (RUBY_DATA_FUNC)rxml_node_free;
+  xnode->_private = (void*)node;
+}
 
-  if (xnode->doc != NULL)
-    rb_gc_mark((VALUE) xnode->doc->_private);
-  
-  if (xnode->parent != NULL)
-    rb_gc_mark((VALUE) xnode->_private);
+void rxml_node_unmanage(xmlNodePtr xnode, VALUE node)
+{
+  RDATA(node)->dfree = NULL;
+  xnode->_private = NULL;
+}
+
+xmlNodePtr rxml_node_root(xmlNodePtr xnode)
+{
+  xmlNodePtr current = xnode;
+
+  while (current->parent)
+  {
+    current = current->parent;
+  }
+
+  return current;
+}
+
+void rxml_node_mark(xmlNodePtr xnode)
+{
+   if (xnode->doc)
+   {
+     if (xnode->doc->_private)
+	 {
+	   VALUE doc = (VALUE)xnode->doc->_private;
+	   rb_gc_mark(doc);
+	 }
+   }
+   else if (xnode->parent)
+   {
+     xmlNodePtr root = rxml_node_root(xnode);
+     if (root->_private)
+     {
+       VALUE node = (VALUE)root->_private;
+       rb_gc_mark(node);
+     }
+   }
 }
 
 VALUE rxml_node_wrap(xmlNodePtr xnode)
 {
-  /* Is the node already wrapped? */
-  if (xnode->_private != NULL)
+  VALUE result = Qnil;
+
+  // Is this node already wrapped?
+  if (xnode->_private)
   {
-    return (VALUE) xnode->_private;
+    result = (VALUE)xnode->_private;
   }
   else
   {
-    VALUE node = Data_Wrap_Struct(cXMLNode, rxml_node_mark, rxml_node_free, xnode);
-    xnode->_private = (void*) node;
-    return node;
+    result = Data_Wrap_Struct(cXMLNode, rxml_node_mark, NULL, xnode);
   }
+
+  if (!xnode->doc && !xnode->parent)
+  {
+    rxml_node_manage(xnode, result);
+  }
+  return result;
 }
 
 static VALUE rxml_node_alloc(VALUE klass)
 {
-  /* Ruby is responsible for freeing this node not libxml but don't set
-     up mark and free yet until we assign the node. */
+  // This node was created from Ruby so we are responsible for freeing it not libxml
   return Data_Wrap_Struct(klass, rxml_node_mark, rxml_node_free, NULL);
 }
 
+static xmlNodePtr rxml_get_xnode(VALUE node)
+{
+   xmlNodePtr result;
+   Data_Get_Struct(node, xmlNode, result);
+
+   if (!result)
+    rb_raise(rb_eRuntimeError, "This node has already been freed.");
+
+   return result;
+}
+
 /*
  * call-seq:
  *    XML::Node.new_cdata(content = nil) -> XML::Node
@@ -128,8 +163,7 @@ static VALUE rxml_node_new_cdata(int argc, VALUE *argv, VALUE klass)
   else
   {
     content = rb_obj_as_string(content);
-    xnode = xmlNewCDataBlock(NULL, (xmlChar*) StringValuePtr(content),
-        RSTRING_LEN(content));
+    xnode = xmlNewCDataBlock(NULL, (xmlChar*) StringValuePtr(content), (int)RSTRING_LEN(content));
   }
 
   if (xnode == NULL)
@@ -169,6 +203,43 @@ static VALUE rxml_node_new_comment(int argc, VALUE *argv, VALUE klass)
   return rxml_node_wrap(xnode);
 }
 
+/*
+ * call-seq:
+ *    XML::Node.new_pi(name, content = nil) -> XML::Node
+ *
+ * Create a new pi node, optionally setting
+ * the node's content.
+ *
+ */
+static VALUE rxml_node_new_pi(int argc, VALUE *argv, VALUE klass)
+{
+  VALUE name = Qnil;
+  VALUE content = Qnil;
+  xmlNodePtr xnode;
+
+  rb_scan_args(argc, argv, "11", &name, &content);
+
+  if (NIL_P(name))
+  {
+    rb_raise(rb_eRuntimeError, "You must provide me with a name for a PI.");
+  }
+  name = rb_obj_as_string(name);
+  if (NIL_P(content))
+  {
+    xnode = xmlNewPI((xmlChar*) StringValuePtr(name), NULL);
+  }
+  else
+  {
+    content = rb_obj_as_string(content);
+    xnode = xmlNewPI((xmlChar*) StringValuePtr(name), (xmlChar*) StringValueCStr(content));
+  }
+
+  if (xnode == NULL)
+    rxml_raise(&xmlLastError);
+
+  return rxml_node_wrap(xnode);
+}
+
 /*
  * call-seq:
  *    XML::Node.new_text(content) -> XML::Node
@@ -219,9 +290,11 @@ static VALUE rxml_node_initialize(int argc, VALUE *argv, VALUE self)
   if (xnode == NULL)
     rxml_raise(&xmlLastError);
 
-  /* Link the Ruby object to the libxml object and vice-versa. */
-  xnode->_private = (void*) self;
-  DATA_PTR(self) = xnode;
+  // Link the ruby wrapper to the underlying libxml node
+  RDATA(self)->data = xnode;
+
+  // Ruby is in charge of managing this node's memory
+  rxml_node_manage(xnode, self);
 
   if (!NIL_P(content))
     rxml_node_content_set(self, content);
@@ -237,24 +310,26 @@ static VALUE rxml_node_modify_dom(VALUE self, VALUE target,
   if (rb_obj_is_kind_of(target, cXMLNode) == Qfalse)
     rb_raise(rb_eTypeError, "Must pass an XML::Node object");
 
-  Data_Get_Struct(self, xmlNode, xnode);
-  Data_Get_Struct(target, xmlNode, xtarget);
+  xnode = rxml_get_xnode(self);
+  xtarget = rxml_get_xnode(target);
 
   if (xtarget->doc != NULL && xtarget->doc != xnode->doc)
-    rb_raise(eXMLError, "Nodes belong to different documents.  You must first import the by calling XML::Document.import");
+    rb_raise(eXMLError, "Nodes belong to different documents.  You must first import the node by calling LibXML::XML::Document.import");
 
-  /* This target node could be freed here. */  
+  xmlUnlinkNode(xtarget);
+
+  // Target is about to have a parent, so stop having ruby manage it.
+  rxml_node_unmanage(xtarget, target);
+
+  // This target node could be freed here and be replaced by a different node 
   xresult = xmlFunc(xnode, xtarget);
 
   if (!xresult)
-    rxml_raise(&xmlLastError);
+	  rxml_raise(&xmlLastError);
 
-  /* Was the target freed? If yes, then wrap the new node */
-  if (xresult != xtarget)
-  {
-    RDATA(target)->data = xresult;
-    xresult->_private = (void*) target;
-  }
+  /* Assume the target was freed, we need to fix up the ruby object to point to the
+     newly returned node. */
+  RDATA(target)->data = xresult;
 
   return target;
 }
@@ -271,7 +346,7 @@ static VALUE rxml_node_base_uri_get(VALUE self)
   xmlChar* base_uri;
   VALUE result = Qnil;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   if (xnode->doc == NULL)
     return (result);
@@ -279,7 +354,7 @@ static VALUE rxml_node_base_uri_get(VALUE self)
   base_uri = xmlNodeGetBase(xnode->doc, xnode);
   if (base_uri)
   {
-    result = rb_str_new2((const char*) base_uri);
+    result = rxml_new_cstr( base_uri, NULL);
     xmlFree(base_uri);
   }
 
@@ -299,7 +374,7 @@ static VALUE rxml_node_base_uri_set(VALUE self, VALUE uri)
   xmlNodePtr xnode;
 
   Check_Type(uri, T_STRING);
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   if (xnode->doc == NULL)
     return (Qnil);
 
@@ -319,11 +394,11 @@ static VALUE rxml_node_content_get(VALUE self)
   xmlChar *content;
   VALUE result = Qnil;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   content = xmlNodeGetContent(xnode);
   if (content)
   {
-    result = rb_str_new2((const char *) content);
+    result = rxml_new_cstr(content, NULL);
     xmlFree(content);
   }
 
@@ -339,43 +414,16 @@ static VALUE rxml_node_content_get(VALUE self)
 static VALUE rxml_node_content_set(VALUE self, VALUE content)
 {
   xmlNodePtr xnode;
+  xmlChar* encoded_content;
 
   Check_Type(content, T_STRING);
-  Data_Get_Struct(self, xmlNode, xnode);
-  // XXX docs indicate need for escaping entites, need to be done? danj
-  xmlNodeSetContent(xnode, (xmlChar*) StringValuePtr(content));
+  xnode = rxml_get_xnode(self);
+  encoded_content = xmlEncodeSpecialChars(xnode->doc, (xmlChar*) StringValuePtr(content));
+  xmlNodeSetContent(xnode, encoded_content);
+  xmlFree(encoded_content);
   return (Qtrue);
 }
 
-/*
- * call-seq:
- *    node.content_stripped -> "string"
- *
- * Obtain this node's stripped content.
- *
- * *Deprecated*: Stripped content can be obtained via the
- * +content+ method.
- */
-static VALUE rxml_node_content_stripped_get(VALUE self)
-{
-  xmlNodePtr xnode;
-  xmlChar* content;
-  VALUE result = Qnil;
-
-  Data_Get_Struct(self, xmlNode, xnode);
-
-  if (!xnode->content)
-    return result;
-
-  content = xmlNodeGetContent(xnode);
-  if (content)
-  {
-    result = rb_str_new2((const char*) content);
-    xmlFree(content);
-  }
-  return (result);
-}
-
 /*
  * call-seq:
  *    node.debug -> true|false
@@ -387,11 +435,11 @@ static VALUE rxml_node_debug(VALUE self)
 {
 #ifdef LIBXML_DEBUG_ENABLED
   xmlNodePtr xnode;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   xmlDebugDumpNode(NULL, xnode, 2);
   return Qtrue;
 #else
-  rb_warn("libxml was compiled without debugging support.")
+  rb_warn("libxml was compiled without debugging support.");
   return Qfalse;
 #endif
 }
@@ -406,7 +454,7 @@ static VALUE rxml_node_first_get(VALUE self)
 {
   xmlNodePtr xnode;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   if (xnode->children)
     return (rxml_node_wrap(xnode->children));
@@ -417,7 +465,7 @@ static VALUE rxml_node_first_get(VALUE self)
 
 /*
  * call-seq:
- *   curr_node << "<p>A paragraph</p>" 
+ *   curr_node << "Some text" 
  *   curr_node << node
  *
  * Add  the specified text or XML::Node as a new child node to the 
@@ -434,16 +482,14 @@ static VALUE rxml_node_content_add(VALUE self, VALUE obj)
   xmlNodePtr xnode;
   VALUE str;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
+
   /* XXX This should only be legal for a CDATA type node, I think,
    * resulting in a merge of content, as if a string were passed
    * danj 070827
    */
   if (rb_obj_is_kind_of(obj, cXMLNode))
   { 
-    xmlNodePtr xtarget;
-    Data_Get_Struct(obj, xmlNode, xtarget);
-    xmlUnlinkNode(xtarget);
     rxml_node_modify_dom(self, obj, xmlAddChild);
   }
   else
@@ -465,41 +511,29 @@ static VALUE rxml_node_content_add(VALUE self, VALUE obj)
  */
 static VALUE rxml_node_doc(VALUE self)
 {
-  xmlNodePtr xnode;
-  xmlDocPtr doc = NULL;
-
-  Data_Get_Struct(self, xmlNode, xnode);
+  xmlDocPtr xdoc = NULL;
+  xmlNodePtr xnode = rxml_get_xnode(self);
 
   switch (xnode->type)
   {
   case XML_DOCUMENT_NODE:
 #ifdef LIBXML_DOCB_ENABLED
-    case XML_DOCB_DOCUMENT_NODE:
+  case XML_DOCB_DOCUMENT_NODE:
 #endif
   case XML_HTML_DOCUMENT_NODE:
-    doc = NULL;
+  case XML_NAMESPACE_DECL:
     break;
   case XML_ATTRIBUTE_NODE:
-  {
-    xmlAttrPtr attr = (xmlAttrPtr) xnode;
-    doc = attr->doc;
-    break;
-  }
-  case XML_NAMESPACE_DECL:
-    doc = NULL;
+    xdoc = (xmlDocPtr)((xmlAttrPtr) xnode->doc);
     break;
   default:
-    doc = xnode->doc;
-    break;
+    xdoc = xnode->doc;
   }
 
-  if (doc == NULL)
+  if (xdoc == NULL)
     return (Qnil);
 
-  if (doc->_private == NULL)
-    rb_raise(rb_eRuntimeError, "existing document object has no ruby-instance");
-
-  return (VALUE) doc->_private;
+  return (VALUE)xdoc->_private;
 }
 
 /*
@@ -538,7 +572,7 @@ static VALUE rxml_node_to_s(int argc, VALUE *argv, VALUE self)
 
   int level = 0;
   int indent = 1;
-  const char *xencoding = NULL;
+  const xmlChar *xencoding = (const xmlChar*)"UTF-8";
 
   rb_scan_args(argc, argv, "01", &options);
 
@@ -558,23 +592,31 @@ static VALUE rxml_node_to_s(int argc, VALUE *argv, VALUE self)
 
     if (rencoding != Qnil)
     {
-      xencoding = xmlGetCharEncodingName((xmlCharEncoding)NUM2INT(rencoding));
+      xencoding = (const xmlChar*)xmlGetCharEncodingName((xmlCharEncoding)NUM2INT(rencoding));
       if (!xencoding)
         rb_raise(rb_eArgError, "Unknown encoding value: %d", NUM2INT(rencoding));
     }
   }
 
-  encodingHandler = xmlFindCharEncodingHandler(xencoding);
+  encodingHandler = xmlFindCharEncodingHandler((const char*)xencoding);
   output = xmlAllocOutputBuffer(encodingHandler);
 
-  Data_Get_Struct(self, xmlNode, xnode);
-  xmlNodeDumpOutput(output, xnode->doc, xnode, level, indent, xencoding);
+  xnode = rxml_get_xnode(self);
+
+  xmlNodeDumpOutput(output, xnode->doc, xnode, level, indent, (const char*)xencoding);
   xmlOutputBufferFlush(output);
 
+#ifdef LIBXML2_NEW_BUFFER
   if (output->conv)
-    result = rb_str_new2((const char*) output->conv->content);
+    result = rxml_new_cstr(xmlBufContent(output->conv), xencoding);
   else
-    result = rb_str_new2((const char*) output->buffer->content);
+    result = rxml_new_cstr(xmlBufContent(output->buffer), xencoding);
+#else
+  if (output->conv)
+    result = rxml_new_cstr(xmlBufferContent(output->conv), xencoding);
+  else
+    result = rxml_new_cstr(xmlBufferContent(output->buffer), xencoding);
+#endif
 
   xmlOutputBufferClose(output);
   
@@ -597,7 +639,7 @@ static VALUE rxml_node_each(VALUE self)
 {
   xmlNodePtr xnode;
   xmlNodePtr xcurrent;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   xcurrent = xnode->children;
 
@@ -617,12 +659,12 @@ static VALUE rxml_node_each(VALUE self)
  * call-seq:
  *    node.empty? -> (true|false)
  *
- * Determine whether this node is empty.
+ * Determine whether this node is an empty or whitespace only text-node.
  */
 static VALUE rxml_node_empty_q(VALUE self)
 {
   xmlNodePtr xnode;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   if (xnode == NULL)
     return (Qnil);
 
@@ -635,29 +677,23 @@ static VALUE rxml_node_empty_q(VALUE self)
  *    node.eql?(other_node) => (true|false)
  *
  * Test equality between the two nodes. Two nodes are equal
- * if they are the same node or have the same XML representation.*/
+ * if they are the same node.*/
 static VALUE rxml_node_eql_q(VALUE self, VALUE other)
 {
-if(self == other)
-{
-  return Qtrue;
-}
-else if (NIL_P(other))
-{
-  return Qfalse;
-}
-else
-{
-  VALUE self_xml;
-  VALUE other_xml;
-
-  if (rb_obj_is_kind_of(other, cXMLNode) == Qfalse)
-  rb_raise(rb_eTypeError, "Nodes can only be compared against other nodes");
-
-  self_xml = rxml_node_to_s(0, NULL, self);
-  other_xml = rxml_node_to_s(0, NULL, other);
-  return(rb_funcall(self_xml, rb_intern("=="), 1, other_xml));
-}
+  if (self == other)
+  {
+    return Qtrue;
+  }
+  else if (NIL_P(other))
+  {
+    return Qfalse;
+  }
+  else
+  {
+    xmlNodePtr xnode = rxml_get_xnode(self);
+    xmlNodePtr xnode_other = rxml_get_xnode(other);
+    return xnode == xnode_other ? Qtrue : Qfalse;
+  }
 }
 
 /*
@@ -673,12 +709,12 @@ static VALUE rxml_node_lang_get(VALUE self)
   xmlChar *lang;
   VALUE result = Qnil;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   lang = xmlNodeGetLang(xnode);
 
   if (lang)
   {
-    result = rb_str_new2((const char*) lang);
+    result = rxml_new_cstr( lang, NULL);
     xmlFree(lang);
   }
 
@@ -699,7 +735,7 @@ static VALUE rxml_node_lang_set(VALUE self, VALUE lang)
   xmlNodePtr xnode;
 
   Check_Type(lang, T_STRING);
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   xmlNodeSetLang(xnode, (xmlChar*) StringValuePtr(lang));
 
   return (Qtrue);
@@ -715,7 +751,7 @@ static VALUE rxml_node_last_get(VALUE self)
 {
   xmlNodePtr xnode;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   if (xnode->last)
     return (rxml_node_wrap(xnode->last));
@@ -735,7 +771,7 @@ static VALUE rxml_node_line_num(VALUE self)
 {
   xmlNodePtr xnode;
   long line_num;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   if (!xmlLineNumbersDefaultValue)
     rb_warn(
@@ -759,7 +795,7 @@ static VALUE rxml_node_xlink_q(VALUE self)
   xmlNodePtr xnode;
   xlinkType xlt;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   xlt = xlinkIsLink(xnode->doc, xnode);
 
   if (xlt == XLINK_TYPE_NONE)
@@ -781,7 +817,7 @@ static VALUE rxml_node_xlink_type(VALUE self)
   xmlNodePtr xnode;
   xlinkType xlt;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   xlt = xlinkIsLink(xnode->doc, xnode);
 
   if (xlt == XLINK_TYPE_NONE)
@@ -803,7 +839,7 @@ static VALUE rxml_node_xlink_type_name(VALUE self)
   xmlNodePtr xnode;
   xlinkType xlt;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   xlt = xlinkIsLink(xnode->doc, xnode);
 
   switch (xlt)
@@ -811,11 +847,11 @@ static VALUE rxml_node_xlink_type_name(VALUE self)
   case XLINK_TYPE_NONE:
     return (Qnil);
   case XLINK_TYPE_SIMPLE:
-    return (rb_str_new2("simple"));
+    return (rxml_new_cstr((const xmlChar*)"simple", NULL));
   case XLINK_TYPE_EXTENDED:
-    return (rb_str_new2("extended"));
+    return (rxml_new_cstr((const xmlChar*)"extended", NULL));
   case XLINK_TYPE_EXTENDED_SET:
-    return (rb_str_new2("extended_set"));
+    return (rxml_new_cstr((const xmlChar*)"extended_set", NULL));
   default:
     rb_fatal("Unknowng xlink type, %d", xlt);
   }
@@ -832,7 +868,7 @@ static VALUE rxml_node_name_get(VALUE self)
   xmlNodePtr xnode;
   const xmlChar *name;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   switch (xnode->type)
   {
@@ -866,7 +902,7 @@ static VALUE rxml_node_name_get(VALUE self)
   if (xnode->name == NULL)
     return (Qnil);
   else
-    return (rb_str_new2((const char*) name));
+    return (rxml_new_cstr( name, NULL));
 }
 
 /*
@@ -881,7 +917,7 @@ static VALUE rxml_node_name_set(VALUE self, VALUE name)
   const xmlChar *xname;
 
   Check_Type(name, T_STRING);
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   xname = (const xmlChar*)StringValuePtr(name);
 
 	/* Note: calling xmlNodeSetName() for a text node is ignored by libXML. */
@@ -900,7 +936,7 @@ static VALUE rxml_node_next_get(VALUE self)
 {
   xmlNodePtr xnode;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   if (xnode->next)
     return (rxml_node_wrap(xnode->next));
@@ -933,7 +969,7 @@ static VALUE rxml_node_parent_get(VALUE self)
 {
   xmlNodePtr xnode;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   if (xnode->parent)
     return (rxml_node_wrap(xnode->parent));
@@ -951,25 +987,18 @@ static VALUE rxml_node_path(VALUE self)
 {
   xmlNodePtr xnode;
   xmlChar *path;
+  VALUE result = Qnil;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   path = xmlGetNodePath(xnode);
 
-  if (path == NULL)
-    return (Qnil);
-  else
-    return (rb_str_new2((const char*) path));
-}
+  if (path)
+  {
+    result = rxml_new_cstr( path, NULL);
+    xmlFree(path);
+  }
 
-/*
- * call-seq:
- *    node.pointer -> XML::NodeSet
- *
- * Evaluates an XPointer expression relative to this node.
- */
-static VALUE rxml_node_pointer(VALUE self, VALUE xptr_str)
-{
-  return (rxml_xpointer_point2(self, xptr_str));
+  return result;
 }
 
 /*
@@ -982,7 +1011,7 @@ static VALUE rxml_node_prev_get(VALUE self)
 {
   xmlNodePtr xnode;
   xmlNodePtr node;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   switch (xnode->type)
   {
@@ -1036,7 +1065,7 @@ static VALUE rxml_node_attributes_get(VALUE self)
 {
   xmlNodePtr xnode;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   return rxml_attributes_new(xnode);
 }
 
@@ -1045,7 +1074,7 @@ static VALUE rxml_node_attributes_get(VALUE self)
  *    node.property("name") -> "string"
  *    node["name"]          -> "string"
  *
- * Obtain the named pyroperty.
+ * Obtain the named property.
  */
 static VALUE rxml_node_attribute_get(VALUE self, VALUE name)
 {
@@ -1076,29 +1105,15 @@ static VALUE rxml_node_property_set(VALUE self, VALUE name, VALUE value)
 
 static VALUE rxml_node_remove_ex(VALUE self)
 {
-  xmlNodePtr xnode, xresult;
-  Data_Get_Struct(self, xmlNode, xnode);
-
-  /* First unlink the node from its parent. */
+  xmlNodePtr xnode = rxml_get_xnode(self);
+ 
+  // Now unlink the node from its parent
   xmlUnlinkNode(xnode);
 
-  /* Now copy the node we want to remove and make the
-     current Ruby object point to it.  We do this because
-     a node has a number of dependencies on its parent
-     document - its name (if using a dictionary), entities,
-     namespaces, etc.  For a node to live on its own, it
-     needs to get its own copies of this information.*/
-  xresult = xmlDocCopyNode(xnode, NULL, 1);
-  
-  /* Now free the original node. */
-  xmlFreeNode(xnode);
+  // Ruby now manages this node
+  rxml_node_manage(xnode, self);
 
-  /* Now wrap the new node */
-  RDATA(self)->data = xresult;
-  xresult->_private = (void*) self;
-
-  /* Now return the removed node so the user can
-     do something with it.*/
+  // Now return the removed node so the user can do something with it
   return self;
 }
 
@@ -1137,7 +1152,7 @@ static VALUE rxml_node_sibling_set(VALUE self, VALUE sibling)
 static VALUE rxml_node_output_escaping_q(VALUE self)
 {
   xmlNodePtr xnode;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   switch (xnode->type) {
   case XML_TEXT_NODE:
@@ -1181,19 +1196,19 @@ static VALUE rxml_node_output_escaping_q(VALUE self)
  * Text nodes which are added to an element or attribute node will be affected
  * by any previous setting of this property.
  */
-static VALUE rxml_node_output_escaping_set(VALUE self, VALUE bool)
+static VALUE rxml_node_output_escaping_set(VALUE self, VALUE value)
 {
   xmlNodePtr xnode;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   switch (xnode->type) {
   case XML_TEXT_NODE:
-    xnode->name = (bool!=Qfalse && bool!=Qnil) ? xmlStringText : xmlStringTextNoenc;
+    xnode->name = (value != Qfalse && value != Qnil) ? xmlStringText : xmlStringTextNoenc;
     break;
   case XML_ELEMENT_NODE:
   case XML_ATTRIBUTE_NODE:
     {
-      const xmlChar *name = (bool!=Qfalse && bool!=Qnil) ? xmlStringText : xmlStringTextNoenc;
+      const xmlChar *name = (value != Qfalse && value != Qnil) ? xmlStringText : xmlStringTextNoenc;
       xmlNodePtr tmp;
       for (tmp = xnode->children; tmp; tmp = tmp->next)
         if (tmp->type == XML_TEXT_NODE)
@@ -1204,7 +1219,7 @@ static VALUE rxml_node_output_escaping_set(VALUE self, VALUE bool)
     return Qnil;
   }
 
-  return (bool!=Qfalse && bool!=Qnil) ? Qtrue : Qfalse;
+  return (value!=Qfalse && value!=Qnil) ? Qtrue : Qfalse;
 }
 
 /*
@@ -1217,7 +1232,7 @@ static VALUE rxml_node_space_preserve_get(VALUE self)
 {
   xmlNodePtr xnode;
 
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   return (INT2NUM(xmlNodeGetSpacePreserve(xnode)));
 }
 
@@ -1227,15 +1242,15 @@ static VALUE rxml_node_space_preserve_get(VALUE self)
  *
  * Control whether this node preserves whitespace.
  */
-static VALUE rxml_node_space_preserve_set(VALUE self, VALUE bool)
+static VALUE rxml_node_space_preserve_set(VALUE self, VALUE value)
 {
   xmlNodePtr xnode;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
-  if (TYPE(bool) == T_FALSE)
-    xmlNodeSetSpacePreserve(xnode, 1);
-  else
+  if (value == Qfalse)
     xmlNodeSetSpacePreserve(xnode, 0);
+  else
+    xmlNodeSetSpacePreserve(xnode, 1);
 
   return (Qnil);
 }
@@ -1249,7 +1264,7 @@ static VALUE rxml_node_space_preserve_set(VALUE self, VALUE bool)
 static VALUE rxml_node_type(VALUE self)
 {
   xmlNodePtr xnode;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
   return (INT2NUM(xnode->type));
 }
 
@@ -1268,7 +1283,7 @@ static VALUE rxml_node_copy(VALUE self, VALUE deep)
   xmlNodePtr xnode;
   xmlNodePtr xcopy;
   int recursive = (deep == Qnil || deep == Qfalse) ? 0 : 1;
-  Data_Get_Struct(self, xmlNode, xnode);
+  xnode = rxml_get_xnode(self);
 
   xcopy = xmlCopyNode(xnode, recursive);
 
@@ -1280,8 +1295,6 @@ static VALUE rxml_node_copy(VALUE self, VALUE deep)
 
 void rxml_init_node(void)
 {
-  xmlDeregisterNodeDefault(rxml_node_deregisterNode);
-
   cXMLNode = rb_define_class_under(mXML, "Node", rb_cObject);
 
   rb_define_const(cXMLNode, "SPACE_DEFAULT", INT2NUM(0));
@@ -1328,6 +1341,7 @@ void rxml_init_node(void)
 
   rb_define_singleton_method(cXMLNode, "new_cdata", rxml_node_new_cdata, -1);
   rb_define_singleton_method(cXMLNode, "new_comment", rxml_node_new_comment, -1);
+  rb_define_singleton_method(cXMLNode, "new_pi", rxml_node_new_pi, -1);
   rb_define_singleton_method(cXMLNode, "new_text", rxml_node_new_text, 1);
 
   /* Initialization */
@@ -1359,7 +1373,6 @@ void rxml_init_node(void)
   rb_define_method(cXMLNode, "copy", rxml_node_copy, 1);
   rb_define_method(cXMLNode, "content", rxml_node_content_get, 0);
   rb_define_method(cXMLNode, "content=", rxml_node_content_set, 1);
-  rb_define_method(cXMLNode, "content_stripped", rxml_node_content_stripped_get, 0);
   rb_define_method(cXMLNode, "debug", rxml_node_debug, 0);
   rb_define_method(cXMLNode, "doc", rxml_node_doc, 0);
   rb_define_method(cXMLNode, "empty?", rxml_node_empty_q, 0);
@@ -1373,7 +1386,6 @@ void rxml_init_node(void)
   rb_define_method(cXMLNode, "output_escaping?", rxml_node_output_escaping_q, 0);
   rb_define_method(cXMLNode, "output_escaping=", rxml_node_output_escaping_set, 1);
   rb_define_method(cXMLNode, "path", rxml_node_path, 0);
-  rb_define_method(cXMLNode, "pointer", rxml_node_pointer, 1);
   rb_define_method(cXMLNode, "remove!", rxml_node_remove_ex, 0);
   rb_define_method(cXMLNode, "space_preserve", rxml_node_space_preserve_get, 0);
   rb_define_method(cXMLNode, "space_preserve=", rxml_node_space_preserve_set, 1);
